/*
 * recovery.c - NILFS recovery logic
 *
 * Copyright (C) 2005, 2006 Nippon Telegraph and Telephone Corporation.
 *
 * This file is part of NILFS.
 *
 * NILFS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NILFS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NILFS; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * recovery.c,v 1.2 2006/04/28 12:22:21 ryusuke Exp
 *
 * Written by Ryusuke Konishi <ryusuke@osrg.net>
 */

#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/crc32c.h>
#include "nilfs.h"

static int
load_segment_summary(struct nilfs_sb_info *sbi, dbn_t pseg_start,
		     struct nilfs_segsum_info *ssi)
{
	struct buffer_head *bh_sum;
	struct nilfs_seg_summary *sum;
	unsigned int magic;
	unsigned long sum_bytes, size;
	unsigned long blocksize = sbi->s_super->s_blocksize;
	int err = NILFS_SEG_FAIL_IO;
	u32 crc, sumsum;

	bh_sum = nilfs_bread(sbi->s_super, pseg_start);
	if(!bh_sum)
		return err;

	sum = (struct nilfs_seg_summary *)bh_sum->b_data;
	magic = le16_to_cpu(sum->ss_magic);
	if(magic != NILFS_SEGSUM_MAGIC) {
		err = NILFS_SEG_FAIL_MAGIC;
		goto fail;
	}

	ssi->nblocks = le16_to_cpu(sum->ss_nblocks);
	ssi->nfinfo = le16_to_cpu(sum->ss_nfinfo);
	ssi->nfbinfo = le16_to_cpu(sum->ss_nfbinfo);
	ssi->nblk_file = le16_to_cpu(sum->ss_nfblk);
	ssi->nblk_fbt = le16_to_cpu(sum->ss_nfbblk);
	ssi->nblk_inode = le16_to_cpu(sum->ss_niblk);
	ssi->nblk_ibt = le16_to_cpu(sum->ss_nibblk);
	ssi->flags = le16_to_cpu(sum->ss_flags);
	ssi->ctime = le64_to_cpu(sum->ss_create);

	sum_bytes = nilfs_recalc_segsum_size(ssi, blocksize);
	if(ssi->nblk_sum > ssi->nblocks) {
		err = NILFS_SEG_FAIL_CONSISTENCY;
		goto fail;
	}
	ssi->prev_pseg = le16_to_cpu(sum->ss_prev);
	if(ssi->prev_pseg >= sbi->s_blocks_per_segment) {
		err = NILFS_SEG_FAIL_CONSISTENCY;
		goto fail;
	}
	ssi->seg_seq = le64_to_cpu(sum->ss_seq);

	/*
	 * Code for checking consistency of the segment summary should be
	 * inserted here.
	 */

	sumsum = le32_to_cpu(sum->ss_sumsum);

	size = min(sum_bytes, blocksize);
	crc = crc32c_le(sbi->s_crc_seed, (unsigned char *)sum +
			sizeof(sum->ss_datasum) + sizeof(sum->ss_sumsum), 
			size - (sizeof(sum->ss_datasum) + sizeof(sum->ss_sumsum)));
	if(ssi->nblk_sum > 1) {
		unsigned int b;
		struct buffer_head *bh;

		for(b = 1; b < ssi->nblk_sum; b++) {
			bh = nilfs_bread(sbi->s_super, pseg_start + b);
			if(!bh)
				goto fail;
			sum_bytes -= size;
			size = min(sum_bytes, blocksize);
			crc = crc32c_le(crc, bh->b_data, size);
			nilfs_brelse(bh);
		}
	}
	if(sumsum != crc) {
		err = NILFS_SEG_FAIL_CHECKSUM_SEGSUM;
		goto fail;
	}
	err = 0;
 fail:
	nilfs_brelse(bh_sum);
	return err;
}

static int
load_checkpoint(struct super_block *sb, dbn_t cp_start, dbn_t cp_end, 
		struct nilfs_cp_info *cpi)
{
	struct buffer_head *bh_cp;
	struct nilfs_checkpoint *cp;
	unsigned long cp_bytes;
	u32 crc, sum;
	int err;
  
	bh_cp = nilfs_bread(sb, cp_start);
	if(!bh_cp)
		return NILFS_SEG_FAIL_IO;

	cp = (struct nilfs_checkpoint *)bh_cp->b_data;

	sum = le32_to_cpu(cp->cp_sum);
	cpi->flags = le16_to_cpu(cp->cp_flags);
	if(!(cpi->flags & NILFS_CP_MAJOR)) {
		err = NILFS_SEG_HAS_NO_MAJOR_CP;
		goto fail;
	}

	cp_bytes = le16_to_cpu(cp->cp_bytes);
	if(cp_start != cp_end) {
		err = NILFS_SEG_FAIL_CONSISTENCY;
		goto fail;
	}

	crc = crc32c_le(NILFS_SB(sb)->s_crc_seed, (unsigned char *)cp + sizeof(cp->cp_sum),
			cp_bytes - sizeof(cp->cp_sum));
	if(sum != crc) {
		err = NILFS_SEG_FAIL_CHECKSUM_CP;
		goto fail;
	}

	cpi->inode_root = le64_to_cpu(cp->cp_inode_root);
	cpi->inodes_count = le64_to_cpu(cp->cp_inodes_count);
	cpi->blocks_count = le64_to_cpu(cp->cp_blocks_count);
	err = 0;
 fail:
	nilfs_brelse(bh_cp);
	return err;
}

int nilfs_warn_segment_error(struct nilfs_sb_info *sbi, int err)
{
	switch (err) {
	case NILFS_SEG_FAIL_IO:
		printk(KERN_WARNING
		       "NILFS warning: I/O error on loading last segment\n");
		return -EIO;
	case NILFS_SEG_FAIL_MAGIC:
		printk(KERN_WARNING
		       "NILFS warning: Invalid magic number at last segment\n");
		break;
	case NILFS_SEG_FAIL_SEQ:
		printk(KERN_WARNING
		       "NILFS warning: Sequence number mismatch\n");
		break;
	case NILFS_SEG_FAIL_CHECKSUM_SEGSUM:
		printk(KERN_WARNING
		       "NILFS warning: Checksum error in segment summary\n");
		break;
	case NILFS_SEG_FAIL_CHECKSUM_CP:
		printk(KERN_WARNING
		       "NILFS warning: Checksum error in checkpoint\n");
		break;
	case NILFS_SEG_FAIL_CHECKSUM_FULL:
		printk(KERN_WARNING
		       "NILFS warning: Checksum error in whole segment check\n");
		break;
	case NILFS_SEG_FAIL_CONSISTENCY:
		printk(KERN_WARNING
		       "NILFS warning: Inconsistent segment\n");
		break;
	case NILFS_SEG_HAS_NO_MAJOR_CP:
		printk(KERN_WARNING
		       "NILFS warning: No major checkpoint in the last "
		       "segment\n");
		break;
	case NILFS_SEG_HAS_NO_CP:
		printk(KERN_WARNING
		       "NILFS warning: No checkpoint in the last segment\n");
		break;
	case NILFS_SEG_BAD_SNAPSHOT:
		printk(KERN_WARNING
		       "NILFS warning: Bad snapshot segment (block number=%lu)\n",
		       sbi->s_snapshot_pseg);
		break;
	case NILFS_SEG_VALID:
	default:
		BUG();
	}
	return -EINVAL;
}

static int
__nilfs_load_last_segment(struct nilfs_sb_info *sbi, struct the_nilfs *nilfs, dbn_t *inode_root)
{
	struct nilfs_segsum_info ssi;
	struct nilfs_cp_info cpi;
	dbn_t pseg_start;
	dbn_t seg_start, seg_end;   /* Starting/ending DBN of full segment */
	dbn_t cp_start, cp_end;
	dbn_t end_block;
	dbn_t last_cp = 0;  /* Pointers of the latest valid checkpoint */
	time_t last_ctime = 0;
	u64 seg_seq;
	unsigned long next_segnum, segnum;
	unsigned int pseg_span = 0;
	int recovered = 0;
	int retry = 0;
	int err;
	enum {
		LAST_PSEG_ST,     /* Looking the CP pointed by superblock */
		ROLLBACK_REV_ST,  /* Scanning older CP (seeking backward in a segment) */
		ROLLBACK_FWD_ST,  /* Scanning older CP (seeking forward in a segment) */
		ROLLFORWARD_ST,   /* Scanning newer CP (to be revised) */
	};
	int search = LAST_PSEG_ST;   /* search mode */

	if (test_bit(THE_NILFS_LOADED, &nilfs->ns_flags)) {
		spin_lock_irq(&nilfs->ns_lock);
		*inode_root = nilfs->ns_inode_root;
		spin_unlock_irq(&nilfs->ns_lock);
		return 0;
	} 
	*inode_root = 0;

	if (nilfs_test_opt(sbi, RECOVERY))
		retry = NILFS_EIO_RETRY_COUNT;

	/*
	 * Copying context of the segment constructor into work variables.
	 * These variables are used to track the latest checkpoint.
	 * ns_last_pseg, ns_seg_seq, ns_last_segnum are initialized at init_nilfs().
	 */
	pseg_start = nilfs->ns_last_pseg;
	seg_seq = nilfs->ns_seg_seq;
	segnum = nilfs->ns_last_segnum;

	/* calculate range of segment */
	nilfs_get_segment_range(sbi, segnum, &seg_start, &seg_end);
	end_block = seg_end + 1;
	recovery_debug(2, 
		       "looking segment (seg_start=%lu, seg_end=%lu, "
		       "segnum=%lu, seg_seq=%llu)\n",
		       seg_start, seg_end, segnum, seg_seq);
	for(;;) {
		recovery_debug(2, "loading summary (pseg_start=%lu)\n",
				pseg_start);
		/* load segment summary */
		err = load_segment_summary(sbi, pseg_start, &ssi);
		recovery_debug(2, "done load_segment_summary (err=%d)\n", err);
		if (err == NILFS_SEG_FAIL_IO && --retry < 0)
			goto io_failure;
		if (nilfs_test_opt(sbi, RECOVERY))
			retry = NILFS_EIO_RETRY_COUNT;

		if (err)
			goto strayed;

		if (seg_seq != ssi.seg_seq) {
			err = NILFS_SEG_FAIL_SEQ;
			goto strayed;
		}
		if (search == LAST_PSEG_ST) {
			nilfs->ns_pseg_offset = pseg_start + ssi.nblocks - seg_start;
			nilfs->ns_prev_pseg = ssi.nblocks;
		}
		if (pseg_start + ssi.nblocks - 1 > seg_end) {
			err = NILFS_SEG_FAIL_CONSISTENCY;
			goto try_next_pseg;
		}
		if (!NILFS_SEG_HAS_MJCP(&ssi)) {
			err = NILFS_SEG_HAS_NO_MAJOR_CP;
			goto try_next_pseg;
		}

		/* calculate location of checkpoint */
		cp_start = pseg_start + nilfs_seg_blocks(&ssi, 0);
		cp_end = pseg_start + ssi.nblocks - 1;
		if (cp_end < cp_start) {
			err = NILFS_SEG_FAIL_CONSISTENCY;
			goto try_next_pseg;
		}

		/* load checkpoint */
		err = load_checkpoint(sbi->s_super, cp_start, cp_end, &cpi);
		recovery_debug(2, "done load_checkpoint (err=%d)\n", err);
		if (err == NILFS_SEG_FAIL_IO && --retry < 0)
			goto io_failure;
		if (nilfs_test_opt(sbi, RECOVERY))
			retry = NILFS_EIO_RETRY_COUNT;

		if (unlikely(err))
			goto try_next_pseg;

		/* A valid checkpoint was found.  saving the context. */
		atomic_set(&sbi->s_inodes_count, cpi.inodes_count);
		atomic_set(&sbi->s_blocks_count, cpi.blocks_count);
		nilfs->ns_last_pseg = pseg_start;
		nilfs->ns_seg_seq = seg_seq;
		nilfs->ns_last_segnum = segnum;
		*inode_root = cpi.inode_root;
		last_cp = cp_start;
		last_ctime = ssi.ctime;

		if (search != LAST_PSEG_ST)
			recovered = 1;

		/* Decide what to do next. */
		switch (search) {
		case ROLLBACK_REV_ST:
			goto cp_found;
		case ROLLBACK_FWD_ST:
			break;
		case LAST_PSEG_ST:
			if (nilfs->ns_mount_state & NILFS_VALID_FS)
				goto cp_found;
			search = ROLLFORWARD_ST;
		case ROLLFORWARD_ST:
			pseg_start += (pseg_span = ssi.nblocks);
			if (pseg_start < seg_end)
				continue;
			goto feed_segment;
		}

	try_next_pseg:
		/* Standing on a course, or met an inconsistent state */
		recovery_debug(2, "try_next_pseg: search=%d, err=%d\n",
			       search, err);
		switch (search) {
		case LAST_PSEG_ST:
			if (err) {
				err = nilfs_warn_segment_error(sbi, err);
				if (!nilfs_test_opt(sbi, RECOVERY))
					goto failed;
			}
			end_block = pseg_start;
			search = ROLLBACK_REV_ST;
		case ROLLBACK_REV_ST:
			if (ssi.prev_pseg > 0) {
				pseg_start -= ssi.prev_pseg;
				continue;
			}
			break;
		case ROLLBACK_FWD_ST:
			pseg_start += ssi.nblocks;
			if (pseg_start < end_block)
				continue;
			if (*inode_root)
				goto cp_found;
			break;
		case ROLLFORWARD_ST:
			pseg_start += (pseg_span = ssi.nblocks);
			if (pseg_start < seg_end)
				continue;
			nilfs_invalidate_segment(nilfs, seg_start, seg_end);
			goto feed_segment;
		}

	strayed:
		/* Off the trail */
		recovery_debug(2, "strayed: search=%d, err=%d\n", search, err);
		switch (search) {
		case LAST_PSEG_ST:
			nilfs_invalidate_segment(nilfs, seg_start, seg_end);
			/* Write will be resumed from the next full segment */
			if (err) {
				err = nilfs_warn_segment_error(sbi, err);
				if (!nilfs_test_opt(sbi, RECOVERY))
					goto failed;
			}
		case ROLLBACK_REV_ST:
			if (pseg_start > seg_start) {
				end_block = pseg_start;
				pseg_start = seg_start;
				search = ROLLBACK_FWD_ST;
				continue;
			}
			break;
		case ROLLBACK_FWD_ST:
			if (++pseg_start < end_block)
				continue; /* crawling blocks forward */
			if (*inode_root)
				goto cp_found;
			break;
		case ROLLFORWARD_ST:
			if (pseg_start == seg_start)
				goto cp_found;
			/* Set write position */
			nilfs->ns_pseg_offset = pseg_start - seg_start;
			nilfs->ns_prev_pseg = pseg_span;
			goto feed_segment;
		}
		/* Looking to the previous full segment */
		err = nilfs_find_prev_segment(sbi, segnum, &next_segnum);
		if (err) {
			if (*inode_root)
				break;
			if (err == -ESRCH) {
				printk(KERN_ERR
				       "NILFS: no valid segment found\n");
				err = -EINVAL;
			}
			goto failed;
		}
		seg_seq--;
		search = ROLLBACK_FWD_ST;
		goto switch_segment;

	feed_segment:
		/* Looking to the next full segment */
		BUG_ON(search != ROLLFORWARD_ST);
		err = nilfs_find_next_segment(sbi, segnum, &next_segnum);
		if (err)
			break;
		seg_seq++;

	switch_segment:
		/* Moving pointers */
		segnum = next_segnum;
		nilfs_get_segment_range(sbi, segnum, &seg_start, &seg_end);
		end_block = seg_end + 1;
		pseg_start = seg_start;
		recovery_debug(2, "looking segment (seg_start=%lu, "
			       "seg_end=%lu, segnum=%lu, seg_seq=%llu)\n",
			       seg_start, seg_end, segnum, seg_seq);
	}

 cp_found:
	/* Updating pointers relating to the latest checkpoint */
	spin_lock_irq(&nilfs->ns_lock);
	nilfs->ns_inode_root = *inode_root;
	nilfs->ns_last_pseg_with_cp = nilfs->ns_last_pseg;
	nilfs->ns_last_cp = last_cp;
	nilfs->ns_last_ctime = last_ctime;
	spin_unlock_irq(&nilfs->ns_lock);
	set_bit(THE_NILFS_LOADED, &nilfs->ns_flags);

	return recovered;

 io_failure:
	if (*inode_root)
		goto cp_found;
	err = -EIO;
 failed:
	return err;
}

int nilfs_load_last_segment(struct nilfs_sb_info *sbi, struct the_nilfs *nilfs)
{
	struct nilfs_sc_info *sci;
	dbn_t inode_root;
	int err, recovered;

	down_write(&nilfs->ns_sem);   /* Guarantees atomicity of this recovery process */
	recovered = __nilfs_load_last_segment(sbi, nilfs, &inode_root);
	if (recovered < 0) {
		err = recovered;
		goto failed;
	}

	/* End of search; initialize the inode b-tree with the root blknum */
	BUG_ON(inode_root == 0);
	recovery_debug(1, "got btree inode root (dbn=%lu)\n", inode_root);
	err = nilfs_btree_init(&sbi->s_inode_root, NILFS_BTREE_INODE,
			       inode_root, sbi->s_super);
	if (err) {
		recovery_debug(0, "NILFS: error initializing inode B-tree.\n");
		goto failed;
	}
	recovery_debug(1, "cp_found: ns_last_segnum=%lu, ns_seg_seq=%llu, "
		       "ns_last_pseg=%lu (segnum from ns_last_pseg=%lu), "
		       "ns_pseg_offset=%lu\n", 
		       nilfs->ns_last_segnum, nilfs->ns_seg_seq,
		       nilfs->ns_last_pseg,
		       nilfs->ns_last_pseg / sbi->s_blocks_per_segment,
		       nilfs->ns_pseg_offset);

	sci = NILFS_SC(sbi);
	if (sci) {
		sci->sc_ctime = sbi->s_nilfs->ns_last_ctime;
		err = nilfs_segctor_reset(sci);
		if (err) {
			printk(KERN_ERR "NILFS: error enabling segment constructor.\n");

			nilfs_btree_clear(&sbi->s_inode_root);
			goto failed;
		}
	}
	up_write(&nilfs->ns_sem);
	return recovered;
 failed:
	up_write(&nilfs->ns_sem);
	return err;
}

int nilfs_load_last_segment_for_remount(struct nilfs_sb_info *sbi, struct the_nilfs *nilfs)
{
	struct nilfs_sc_info *sci;
	dbn_t inode_root, last_pseg_with_cp;
	int err, recovered;

	down_write(&nilfs->ns_sem);   /* Guarantees atomicity of this recovery process */
	
	recovered = __nilfs_load_last_segment(sbi, nilfs, &inode_root);
	if (recovered < 0) {
		printk(KERN_ERR "NILFS (device %s): error loading the latest segment.\n",
		       sbi->s_super->s_id);
		err = recovered;
		goto failed;
	}
	spin_lock_irq(&nilfs->ns_lock);
	last_pseg_with_cp = nilfs->ns_last_pseg_with_cp;
	spin_unlock_irq(&nilfs->ns_lock);

	if (inode_root != NILFS_BTREE_ROOT_DBN(&sbi->s_inode_root) ||
	    last_pseg_with_cp != sbi->s_snapshot_pseg) {
		printk(KERN_WARNING "NILFS (device %s): couldn't "
		       "remount because the current RO-mount is not the latest one.\n",
		       sbi->s_super->s_id);
		err = -EINVAL;
		goto failed;
	}

	sci = NILFS_SC(sbi);
	BUG_ON(!sci);
	sci->sc_ctime = sbi->s_nilfs->ns_last_ctime;
	err = nilfs_segctor_reset(sci);
	if (err) {
		printk(KERN_ERR "NILFS (device %s): error enabling segment constructor.\n",
		       sbi->s_super->s_id);
		goto failed;
	}
	return 0;
 failed:
	up_write(&nilfs->ns_sem);
	return err;
}


int nilfs_load_snapshot(struct nilfs_sb_info *sbi)
{
	struct nilfs_segsum_info ssi;
	struct nilfs_cp_info cpi;
	dbn_t seg_start, seg_end; /* start and end DBN of full segment */
	dbn_t cp_start, cp_end;
	unsigned long snapshot_segnum;
	int err;

	snapshot_segnum = sbi->s_snapshot_pseg / sbi->s_blocks_per_segment;
	if (snapshot_segnum >= sbi->s_nsegment) {
		err = NILFS_SEG_BAD_SNAPSHOT;
		goto fail;
	}
	err = load_segment_summary(sbi, sbi->s_snapshot_pseg, &ssi);
	if (err) 
		goto fail;

	nilfs_get_segment_range(sbi, snapshot_segnum, &seg_start, &seg_end);
	if (sbi->s_snapshot_pseg + ssi.nblocks - 1 > seg_end) {
		err = NILFS_SEG_FAIL_CONSISTENCY;
		goto fail;
	}

	cp_start = sbi->s_snapshot_cp = sbi->s_snapshot_pseg + nilfs_seg_blocks(&ssi, 0);
	cp_end = sbi->s_snapshot_pseg + ssi.nblocks - 1;

	if (cp_end < cp_start) {
		err = ((cp_start == cp_end + 1) ?
		       NILFS_SEG_HAS_NO_CP : NILFS_SEG_FAIL_CONSISTENCY);
		goto fail;
	}

	err = load_checkpoint(sbi->s_super, cp_start, cp_end, &cpi);
	if (err)
		goto fail;

	atomic_set(&sbi->s_inodes_count, cpi.inodes_count);
	atomic_set(&sbi->s_blocks_count, cpi.blocks_count);

	err = nilfs_btree_init(&sbi->s_inode_root, NILFS_BTREE_INODE, cpi.inode_root,
			       sbi->s_super);
	if (err < 0)
		goto fail;

	recovery_debug(1, "got btree inode root (dbn=%lu)\n",
		       cpi.inode_root);
	err = 0;
 fail:
	return err;
}

/* Local Variables:	*/
/* eval: (c-set-style "linux")	*/
/* End:			*/
